/**
 * The MIT License (MIT)
 *
 * Copyright (C) 2013 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.oracle.chess.client.htmljava;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.java.html.json.ComputedProperty;
import net.java.html.json.Function;
import net.java.html.json.Model;
import net.java.html.json.ModelOperation;
import net.java.html.json.OnPropertyChange;
import net.java.html.json.OnReceive;
import net.java.html.json.Property;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
@Model(className = "UI", properties = {
    @Property(name = "status", type = String.class),
    @Property(name = "selectedGameId", type = String.class),
    @Property(name = "boards", type = Board.class, array = true),
    @Property(name = "viewGames", type = Games.class),
    @Property(name = "settings", type = Settings.class),
    @Property(name = "connected", type = boolean.class),
    @Property(name = "disconnected", type = boolean.class),
})
public class UIModel {
    private static final Logger LOG = Logger.getLogger(UIModel.class.getName());
    
    @ComputedProperty static boolean settingsActive(String selectedGameId) {
        return selectedGameId == null;
    }
    
    @ComputedProperty static boolean viewGamesActive(String selectedGameId, boolean connected) {
        return selectedGameId == null && connected;
    }
    
    @ComputedProperty static Board selectedBoard(String selectedGameId, List<Board> boards) {
        Board active = null;
        for (Board b : boards) {
            b.setActive(false);
            if (selectedGameId != null && selectedGameId.equals(b.getGameId())) {
                b.setActive(true);
                active = b;
            }
        }
        return active;
    }
    
    @Function static void activateGame(UI m, Board data) {
        m.setSelectedGameId(data.getGameId());
    }

    @Function static void activateSettings(UI m) {
        m.setSelectedGameId(null);
        refreshGames(m);
    }
    
    @Function static void createGame(UI u) {
        u.setStatus("Creating new game...");
        Request r = new Request();
        r.setMsg(MsgType.CreateGame);
        r.setColor(Color.valueOf(u.getViewGames().getSelectedColor()));
        u.sendMsg(r);
    }

    @ModelOperation @Function static void refreshGames(UI u) {
        u.setStatus("Refreshing games...");
        Request r = new Request();
        r.setMsg(MsgType.QueryGames);
        u.sendMsg(r);
    }
    
    @ModelOperation static void sendMsg(UI ui, Request r) {
        final Settings sttngs = ui.getSettings();
        
        String url = sttngs.getUrl();
        r.setUsername(sttngs.getUsername());
        r.setPassword(sttngs.getPassword());
        
        LOG.log(Level.INFO, "Sending {0} to {1}", new Object[]{r, url});
        ui.queryServer(url, r);
    }
    
    
    @OnReceive(data = Request.class, url = "{url}", method = "WebSocket", onError = "wasAnError") 
    static void queryServer(UI ui, Response r) {
        LOG.log(Level.INFO, "Received {0}", r);
        if (r == null) {
            ui.setDisconnected(false);
            verifyLogin(ui);
            return;
        }
        SWITCH: switch (MsgType.forResponse(r.getMsg())) {
            case CheckCredentials: {
                switch (r.getCheck()) {
                    case NOT_REGISTERED:
                    case VALID:
                        ui.setConnected(true);
                        ui.refreshGames();
                        ui.setStatus("Connected.");
                        return;
                    case INVALID:
                        ui.setStatus("Password is invalid");
                        break;
                }
                disconnect(ui);
            }
                
            case QueryGames:
                ui.getViewGames().listGames(r.getGames(), ui.getSettings().getUsername());
                ui.setStatus("");
                break;
            case CreateGame: {
                ui.setStatus("Game " + r + " created");
                final Board b = Rules.createBoard();
                Rules.initBoard(
                    b, r.getBoard().getWhites(),
                    r.getBoard().getBlacks(),
                    r.getTurn()
                );
                b.setGameId(r.getGameId());
                b.setWhitePlayer(r.getWhitePlayer());
                b.setBlackPlayer(r.getBlackPlayer());
                b.setPlayer(ui.getSettings().getUsername());
                b.updateSummary(r.getSummary());
                ui.getBoards().add(b);
                ui.setSelectedGameId(r.getGameId());
                UIModel.refreshGames(ui);
                break;
            }
            case SendMove: {
                if (r.getBoard() == null) {
                    throw new NullPointerException("No board in " + r);
                }
                final Board b = findBoard(ui, r.getGameId());
                if (b == null) {
                    break;
                }
                b.updateSummary(r.getSummary());
                final String errMsg = r.getError() == null ? null : r.getError().getMessage();
                final List<String> whites = r.getBoard().getWhites();
                final List<String> blacks = r.getBoard().getBlacks();
                final Color turn = r.getNextTurn();
                BoardModel.moveResponse(b, errMsg, whites, blacks, turn, r.getAlert());
                break;
            }
            case JoinGame: {
                Board b;
                if ((b = findBoard(ui, r.getGameId())) != null) {
                    b.setWhitePlayer(r.getWhitePlayer());
                    b.setBlackPlayer(r.getBlackPlayer());
                    b.setPlayer(ui.getSettings().getUsername());
                    b.updateSummary(r.getSummary());
                    ui.setSelectedGameId(r.getGameId());
                    break SWITCH;
                }
                ui.setStatus("Joining " + r.getGameId() + " as " + r.getColor());
                b = Rules.createBoard();
                Rules.initBoard(
                    b, r.getBoard().getWhites(),
                    r.getBoard().getBlacks(),
                    r.getTurn()
                );
                b.setGameId(r.getGameId());
                b.setWhitePlayer(r.getWhitePlayer());
                b.setBlackPlayer(r.getBlackPlayer());
                b.setPlayer(ui.getSettings().getUsername());
                b.updateSummary(r.getSummary());
                if (r.getColor() == Color.B) {
                    BoardModel.rotateBoard(b);
                }
                if (r.getMoves() != null) {
                    for (String move : r.getMoves()) {
                        Move m = BoardModel.MoveImpl.valueOf(move);
                        b.getMoves().add(m);
                    }
                } else {
                    b.getMoves().clear();
                }
                ui.setSelectedGameId(b.getGameId());
                ui.getBoards().add(b);
                break;
            }
            case UpdateGame: {
                final Board b = findBoard(ui, r.getGameId());
                if (b == null) {
                    break;
                }
                b.updateSummary(r.getSummary());
                final Move move = BoardModel.MoveImpl.valueOf(r.getFrom() + r.getTo());
                final List<String> whites = r.getBoard().getWhites();
                final List<String> blacks = r.getBoard().getBlacks();
                final Color turn = r.getNextTurn();
                BoardModel.moveUpdate(b, move, whites, blacks, turn, null);
            }
        }
    }


    private static Board findBoard(UI ui, String gameId) {
        for (Board tmp : ui.getBoards()) {
            if (tmp.getGameId().equals(gameId)) {
                return tmp;
            }
        }
        return null;
    }
    
    
    static void wasAnError(UI ui, Exception t) {
        if (t == null) {
            ui.setConnected(false);
            ui.setDisconnected(true);
            ui.setStatus("Disconnected.");
        } else {
            ui.setStatus("Error: " + t.getLocalizedMessage());
        }
    }
    
    @Function static void joinGame(UI u, Game data) {
        Request r = new Request();
        r.setMsg(MsgType.JoinGame);
        r.setObserver(false);
        r.setGameId(data.getGameId());
        u.sendMsg(r);
    }

    @Function static void observeGame(UI u, Game data) {
        Request r = new Request();
        r.setMsg(MsgType.JoinGame);
        r.setObserver(true);
        r.setGameId(data.getGameId());
        u.sendMsg(r);
    }

    private static final Set<UI> ACTIVE = new HashSet<>();
    @OnPropertyChange("connected") static void clearGames(UI m) {
        if (m.isDisconnected()) {
            m.getBoards().clear();
            ACTIVE.remove(m);
        } else {
            ACTIVE.add(m);
        }
    }
    
    static UI findUI(Board b) {
        for (UI ui : ACTIVE) {
            if (ui.getBoards().contains(b)) {
                return ui;
            }
        }
        return null;
    }
    
    @Function static void leave(UI ui, Board data) {
        ui.getBoards().remove(data);
        ui.setSelectedGameId(null);
    }
    
    @Function static void disconnect(UI ui) {
        ui.queryServer(ui.getSettings().getUrl(), null);
        ui.setStatus("Disconnecting...");
    }

    private static void verifyLogin(UI ui) {
        ui.setStatus("Verifying user credentials...");
        Request r = new Request();
        r.setMsg(MsgType.CheckCredentials);
        ui.sendMsg(r);
    }
    
    @Function static void reconnect(UI ui) {
        final Settings s = ui.getSettings();
        ui.setStatus("Connecting to the server...");
        ui.queryServer(s.getUrl(), null);
        s.write(ui);
    }
    
    public static void initialize(String... args) throws URISyntaxException, IOException {
        UI ui = new UI();
        ui.getSettings().read();
        ui.setStatus("Ready.");
        ui.setDisconnected(true);
        ui.applyBindings();
    }

}
